/* CG/solution/parts/cgp07.c */

/****************************************************************
 *                                                              *
 * This file has been written as a sample solution to an        *
 * exercise in a course given at the High Performance           *
 * Computing Centre Stuttgart (HLRS).                           *
 * It is made freely available with the understanding that      *
 * every copy of this file must include this header and that    *
 * HLRS takes no responsibility for the use of the enclosed     *
 * teaching material.                                           *
 *                                                              *
 * This software is partitioned into 16 tasks.                  *
 * Each task can be implemented by one group of the course      *
 * Files: CG/solution/cg_all.c - Solution                       *
 *        CG/skel/cg_01.c      - Skeletons that should be used  *
 *             .. cg_15.c        to implement each task         *
 *        CG/solution/cgp00.c  - Source part of each task       *
 *             ..     cgp16.c                                   *
 *                                                              *
 * Authors: Dr. Rolf Rabenseifner                               *
 *          Felix Triebel                                       *
 *          Michael Speck                                       *
 *          Felix Triebel                                       *
 *          Traugott Streicher                                  *
 *                                                              *
 * Contact: rabenseifner@hlrs.de                                *
 *                                                              *
 ****************************************************************/
                                                           /*EOH*/
/*>>>>>>>>>>>>>>>>>>>>>> begin of task 07 <<<<<<<<<<<<<<<<<<<<<<*/

#ifndef serial
/*------------------------------
 * void Init_Send_Info2(void)
 *
 * TASK: Communicate positions of elements needed from
 *       (or by) other processors.
 *
 * REFERENCE: Slides 19, 20, 23
 *
 * USED GLOBAL DATA:
 *   IN struct halo_info:     Describes position of non-local elements in the halo
 *   IN int recv_num:                      number of processors we receive data from
 *   IN int* recv_ranks:                   ranks of all real neighbors that are sending halo-data
 *   IN neighbor_struct* recv_halo:        Information about data we want to reveive
 *   IN int send_num:                      number of processors we send data to
 *   IN int* send_ranks:                   ranks of all real neighbors to which halo-data must be sent
 *   IN int start:                         begin of the local chunk
 *   IN OUT neighbor_struct* send_halo:    Information about data we have to send
 *                                         (IN send_halo.size)
 * USED FUNCTIONS: APPL: save_malloc
 *                 MPI : MPI_Irecv, MPI_Send, MPI_Waitall
 */
void Init_Send_Info2(void)
{
        int i, rank, j, local_index, index_in_halo_info;
        MPI_Request *req;
        MPI_Status *statuses;
        int src_rank, dest_rank;
        int *orig_pos;
 
        /* allocate scratch memory for req and statuses */ 
        req = (MPI_Request*)save_malloc(sizeof(MPI_Request)*numprocs);
        statuses = (MPI_Status*)save_malloc(sizeof(MPI_Status)*numprocs);

        /* allocate memory (send_halo[rank].size elements) for send_halo[...].I_loc and .scratch */ 
        for (i=0; i<send_num; i++)
            {
                rank = send_ranks[i];
                send_halo[rank].scratch = (double*)save_malloc(sizeof(double)*send_halo[rank].size);
                send_halo[rank].I_loc = (int*)save_malloc(sizeof(int)*send_halo[rank].size);
            }

        /* Communicate halo_info */
        /* non-blocking: receive from all processes halo info */
        for (i=0; i<send_num; i++)
            {
                src_rank = send_ranks[i];
                MPI_Irecv(send_halo[src_rank].I_loc,
                                send_halo[src_rank].size, MPI_INT, src_rank,
                                MSG_TAG1, MPI_COMM_WORLD, &(req[i]));
            }
 
        /* blocking: send */
        for (i=0; i<recv_num; i++)
            {
                dest_rank = recv_ranks[i];
                orig_pos = (int*)save_malloc(sizeof(int)*recv_halo[dest_rank].size);
                for (j=0; j < recv_halo[dest_rank].size; j++)
                {
                        local_index = recv_halo[dest_rank].I_loc[j];
                        index_in_halo_info = local_index - (end1-start);
                        orig_pos[j] = halo_info.I[index_in_halo_info];
                        // sending the global index;
                        // local index is computed in the loop below: ... = ... - start
                }
                MPI_Send(orig_pos,
                           recv_halo[dest_rank].size, MPI_INT,
                           dest_rank, MSG_TAG1, MPI_COMM_WORLD);
                free(orig_pos);
            }
 
        /* Waitall */
        if (send_num > 0)
                MPI_Waitall(send_num, req, statuses);

        /* update send_halo[_].I_loc[_] */
        for (i=0; i<send_num; i++)
            {
                src_rank = send_ranks[i];
                for (j=0; j < send_halo[src_rank].size; j++)
                {
                        send_halo[src_rank].I_loc[j]
                                 = send_halo[src_rank].I_loc[j] - start;
                }
            }
        /* free scratch memory */ 
        free(req); free(statuses);

}
#endif

/*>>>>>>>>>>>>>>>>>>>>>> -end- of task 07 <<<<<<<<<<<<<<<<<<<<<<*/
