/* CG/solution/parts/cgp16.c */

/****************************************************************
 *                                                              *
 * This file has been written as a sample solution to an        *
 * exercise in a course given at the High Performance           *
 * Computing Centre Stuttgart (HLRS).                           *
 * It is made freely available with the understanding that      *
 * every copy of this file must include this header and that    *
 * HLRS takes no responsibility for the use of the enclosed     *
 * teaching material.                                           *
 *                                                              *
 * This software is partitioned into 16 tasks.                  *
 * Each task can be implemented by one group of the course      *
 * Files: CG/solution/cg_all.c - Solution                       *
 *        CG/skel/cg_01.c      - Skeletons that should be used  *
 *             .. cg_15.c        to implement each task         *
 *        CG/solution/cgp00.c  - Source part of each task       *
 *             ..     cgp16.c                                   *
 *                                                              *
 * Authors: Dr. Rolf Rabenseifner                               *
 *          Felix Triebel                                       *
 *          Michael Speck                                       *
 *          Felix Triebel                                       *
 *          Traugott Streicher                                  *
 *                                                              *
 * Contact: rabenseifner@hlrs.de                                *
 *                                                              *
 ****************************************************************/
                                                           /*EOH*/
/*>>>>>>>>>>>>>>>>>>>>>> begin of task 16 <<<<<<<<<<<<<<<<<<<<<<*/

/************************ main program *************************/
int main(int argc, char *argv[])
{
        int num_iter;
        double norm_r, norm_err, maxnorm_err;
        vector_chunk u, e;
        double time_start=0, time_end=0;

#ifdef serial
        numprocs = 1; my_rank = 0;
#else
        MPI_Init(&argc, &argv);
        MPI_Comm_size(MPI_COMM_WORLD, &numprocs);
        MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);
#endif

        m = 4;
        n = 4;
        iter_max = 500;
        epsilon = 1e-6;
        print_level = 1;
        decomp_dims = 1;
        mprt = 11; nprt = 11;
        m_procs=2; n_procs=2;
#       ifndef serial
        {   int dims[2]; dims[0] = 0; dims[1] = 0;
            MPI_Dims_create(numprocs, 2, dims);
            m_procs = dims[0]; n_procs = dims[1];
        }
#       endif


        /* Parse command line options */
        read_options( argc, argv );

        Init_Decomp();
        if (print_level>=4) Print_Decomp();

        /* Init the distribution of the matrix and the vectors */
        Distribution(m*n);

        /* reserve chunk of matrix A */
        alloc_A();
        Init_Laplace_A(&A);
        if (print_level>=4) print_mat("A", &A);

#ifndef serial
        /* !!! determine communication partners for vector x !!! */
        Init_Vec_Halo();
        if (print_level>=4) print_halo();

        Init_Recv_Info();
        if (print_level>=4) print_recv_info();

        Init_Send_Info1();
        Init_Send_Info2();
        if (print_level>=4) print_send_info();
#endif

        /* reserve chunk of vector b and !!! assign border data */
        b = alloc_vec(0);
        Init_Boundary_Vec(b);

        if (print_level>=4) print_vec("b",b);

        /* x is initialized with zeros */
        x = alloc_vec(0);

#ifndef serial
        time_start = MPI_Wtime();
#endif

        /* solves Ax = b,  A,x,b are global variables */
        CG_solve(&num_iter, &norm_r);

#ifndef serial
        time_end = MPI_Wtime();
#endif

        u = alloc_vec(0);
        e = alloc_vec(0);
        Init_Exact_Solution_vec(u);
        /* p = u - x */
        add_vec_vec(u, -1, x, e);
        norm_err = sqrt(sqr_norm_vec(e));
        maxnorm_err = max_norm_vec(e);

        if (print_level>=3) print_vec2d("x",x,m,n);

        if (my_rank==0)
        {
            printf("main: %3d iterations, sqrt(eps) = %8.2e,  normalized: %8.2e\n",num_iter,
                                             sqrt(epsilon), sqrt(epsilon/(n*m)));
            printf("main:                 ||r||     = %8.2e,  normalized: %8.2e\n",
                                             norm_r, norm_r/sqrt(n*m));
            printf("main:                 ||error|| = %8.2e,  normalized: %8.2e\n",
                                             norm_err, norm_err/sqrt(n*m));
            printf("main:                 max(error)= %8.2e = normalized: %8.2e\n",
                                             maxnorm_err, maxnorm_err);
            printf("main:                 n*m = %d*%d = %d\n",n, m, n*m);
            printf("main:                 %8.2e sec\n",time_end-time_start);
        }

        if (print_level>=1) print_vec2d("x",x,mprt,nprt);

#ifndef serial
        MPI_Finalize();
#endif
        return 0;

#ifndef serial
        /* vorzeitiger Test-Stop */ MPI_Finalize();
#endif
        /* vorzeitiger Test-Stop */ return 0;
}
/*>>>>>>>>>>>>>>>>>>>>>> -end- of task 16 <<<<<<<<<<<<<<<<<<<<<<*/
